@@ -11,8 +11,11 @@
 #include <setjmp.h>
 #include <stdio.h>
 #include <sys/socket.h>
+#include <sys/time.h>
 #include <netinet/in.h>
+#include <assert.h>
 
+#include "config.h"
 #if defined (HAVE_PATH_H)
 #include <paths.h>
 #else
@@ -33,8 +36,9 @@
 
 struct gps_session_t *session;
 static char *device_name = DEFAULT_DEVICE_NAME;
+static char *pid_file = NULL;
 static fd_set all_fds, nmea_fds, watcher_fds;
-static int debuglevel, nfds, in_background = 0;
+static int debuglevel, nfds, go_background = 1, in_background = 0;
 
 static jmp_buf	restartbuf;
 #define THROW_SIGHUP	1
@@ -51,6 +55,18 @@ static void onsig(int sig)
     exit(10 + sig);
 }
 
+static void store_pid(pid_t pid)
+{
+    FILE *fp;
+
+    if ((fp = fopen(pid_file, "w")) != NULL) {
+	fprintf(fp, "%u\n", pid);
+	(void) fclose(fp);
+    } else {
+	gpsd_report(1, "Cannot create PID file: %s.\n", pid_file);
+    }
+}
+
 static int daemonize(void)
 {
     int fd;
@@ -59,10 +75,12 @@ static int daemonize(void)
     switch (pid = fork()) {
     case -1:
 	return -1;
-    case 0:
+    case 0:	/* child side */
 	break;
-    default:
-	_exit(pid);
+    default:	/* parent side */
+	if (pid_file)
+		store_pid(pid);
+	exit(0);
     }
 
     if (setsid() == -1)
@@ -92,7 +110,7 @@ void gpsd_report(int errlevel, const char *fmt, ... )
 	va_end(ap);
 
 	if (in_background)
-	    syslog((errlevel == 0) ? LOG_ERR : LOG_NOTICE, buf);
+	    syslog((errlevel == 0) ? LOG_ERR : LOG_NOTICE, "%s", buf);
 	else
 	    fputs(buf, stderr);
     }
@@ -110,8 +128,9 @@ static void usage(void)
 #ifdef TRIPMATE_ENABLE
 "  -i %%f[NS]:%%f[EW]               = set initial latitude/longitude \n"
 #endif /* TRIPMATE_ENABLE */
-"  -s baud_rate                   = set baud rate on gps device \n\
+"  -s baud_rate                   = set baud rate on GPS device \n\
   -d host[:port]                 = set DGPS server \n\
+  -P pidfile                     = set file to record process ID \n\
   -D integer (default 0)         = set debug level \n\
   -h                             = help message \n",
 	   DEFAULT_DEVICE_NAME, DEFAULT_GPSD_PORT);
@@ -167,9 +186,18 @@ static int validate(void)
 static int handle_request(int fd, char *buf, int buflen)
 /* interpret a client request; fd is the socket back to the client */
 {
-    char reply[BUFSIZE], *p;
+    char reply[BUFSIZ], phrase[BUFSIZ], *p;
     int i, j;
     struct gps_data_t *ud = &session->gNMEAdata;
+#ifdef PROFILING
+    int icd = 0;
+#endif /* PROFILING */
+
+#ifdef PROFILING
+    struct timeval tv;
+    gettimeofday(&tv, NULL);
+    session->poll_times[fd] = TIME2DOUBLE(tv);
+#endif /* PROFILING */
 
     sprintf(reply, "GPSD");
     p = buf;
@@ -177,106 +205,130 @@ static int handle_request(int fd, char *buf, int buflen)
 	switch (toupper(*p++)) {
 	case 'A':
 	    if (!validate())
-		strcat(reply, ",A=?");
+		strcpy(phrase, ",A=?");
 	    else
-		sprintf(reply + strlen(reply), ",A=%f", ud->altitude);
+		sprintf(phrase, ",A=%f", ud->altitude);
+	    break;
+	case 'B':		/* change baud rate (SiRF only) */
+#ifdef PROFILING
+	    if (*p == '=') {
+		i = atoi(++p);
+		while (isdigit(*p)) p++;
+		if (session->device_type->speed_switcher)
+		    if (session->device_type->speed_switcher(session, i))
+			gpsd_set_speed(session, (speed_t)i);
+	    }
+#endif /* PROFILING */
+	    sprintf(phrase, ",B=%d %d N %d", 
+		    gpsd_get_speed(&session->ttyset),
+		    9-session->device_type->stopbits,
+		    session->device_type->stopbits);
+	    break;
+	case 'C':
+	    sprintf(phrase, ",C=%d", session->device_type->cycle);
 	    break;
 	case 'D':
-	    if (ud->utc[0])
-		sprintf(reply + strlen(reply), ",D=%s", ud->utc);
-	    else
-		strcat(reply, ",D=?");
+	    if (ud->utc[0]) {
+		sprintf(phrase, ",D=%s", ud->utc);
+#ifdef PROFILING
+		icd = 1;
+#endif /* PROFILING */
+	    } else
+		strcpy(phrase, ",D=?");
 	    break;
 	case 'L':
-	    sprintf(reply + strlen(reply), ",l=1 " VERSION " admpqrstvwxy");
+	    sprintf(phrase, ",l=1 " VERSION " abcdmpqrstvwxy");
 	    break;
 	case 'M':
 	    if (ud->mode == MODE_NOT_SEEN)
-		strcat(reply, ",M=?");
+		strcpy(phrase, ",M=?");
 	    else
-		sprintf(reply + strlen(reply), ",M=%d", ud->mode);
+		sprintf(phrase, ",M=%d", ud->mode);
 	    break;
 	case 'P':
 	    if (!validate())
-		strcat(reply, ",P=?");
+		strcpy(phrase, ",P=?");
 	    else
-		sprintf(reply + strlen(reply), ",P=%f %f", 
+		sprintf(phrase, ",P=%f %f", 
 			ud->latitude, ud->longitude);
 	    break;
 	case 'Q':
 	    if (!validate())
-		strcat(reply, ",Q=?");
+		strcpy(phrase, ",Q=?");
 	    else
-		sprintf(reply + strlen(reply), ",Q=%d %f %f %f",
+		sprintf(phrase, ",Q=%d %f %f %f",
 			ud->satellites_used, ud->pdop, ud->hdop, ud->vdop);
 	    break;
 	case 'R':
+	    if (*p == '=') ++p;
 	    if (*p == '1' || *p == '+') {
 		FD_SET(fd, &nmea_fds);
 		gpsd_report(3, "%d turned on raw mode\n", fd);
-		sprintf(reply + strlen(reply), ",R=1");
+		sprintf(phrase, ",R=1");
 		p++;
 	    } else if (*p == '0' || *p == '-') {
 		FD_CLR(fd, &nmea_fds);
 		gpsd_report(3, "%d turned off raw mode\n", fd);
-		sprintf(reply + strlen(reply), ",R=0");
+		sprintf(phrase, ",R=0");
 		p++;
 	    } else if (FD_ISSET(fd, &nmea_fds)) {
 		FD_CLR(fd, &nmea_fds);
 		gpsd_report(3, "%d turned off raw mode\n", fd);
-		sprintf(reply + strlen(reply), ",R=0");
+		sprintf(phrase, ",R=0");
 	    } else {
 		FD_SET(fd, &nmea_fds);
 		gpsd_report(3, "%d turned on raw mode\n", fd);
-		sprintf(reply + strlen(reply), ",R=1");
+		sprintf(phrase, ",R=1");
 	    }
 	    break;
 	case 'S':
-	    sprintf(reply + strlen(reply), ",S=%d", ud->status);
+	    sprintf(phrase, ",S=%d", ud->status);
 	    break;
 	case 'T':
 	    if (!validate())
-		strcat(reply, ",T=?");
+		strcpy(phrase, ",T=?");
 	    else
-		sprintf(reply + strlen(reply), ",T=%f", ud->track);
+		sprintf(phrase, ",T=%f", ud->track);
 	    break;
 	case 'V':
 	    if (!validate())
-		strcat(reply, ",V=?");
+		strcpy(phrase, ",V=?");
 	    else
-		sprintf(reply + strlen(reply), ",V=%f", ud->speed);
+		sprintf(phrase, ",V=%f", ud->speed);
 	    break;
 	case 'W':
+	    if (*p == '=') ++p;
 	    if (*p == '1' || *p == '+') {
 		FD_SET(fd, &watcher_fds);
 		gpsd_report(3, "%d turned on watching\n", fd);
-		sprintf(reply + strlen(reply), ",W=1");
+		sprintf(phrase, ",W=1");
 		p++;
 	    } else if (*p == '0' || *p == '-') {
 		FD_CLR(fd, &watcher_fds);
 		gpsd_report(3, "%d turned off watching\n", fd);
-		sprintf(reply + strlen(reply), ",W=0");
+		sprintf(phrase, ",W=0");
 		p++;
 	    } else if (FD_ISSET(fd, &watcher_fds)) {
 		FD_CLR(fd, &watcher_fds);
 		gpsd_report(3, "%d turned off watching\n", fd);
-		sprintf(reply + strlen(reply), ",W=0");
+		sprintf(phrase, ",W=0");
 	    } else {
 		FD_SET(fd, &watcher_fds);
 		gpsd_report(3, "%d turned on watching\n", fd);
-		sprintf(reply + strlen(reply), ",W=1");
+		sprintf(phrase, ",W=1");
 	    }
 	    break;
         case 'X':
-	    sprintf(reply + strlen(reply), ",X=%d", ud->online);
+	    sprintf(phrase, ",X=%d", ud->online);
 	    break;
 	case 'Y':
 	    if (!ud->satellites)
-		strcat(reply, ",Y=?");
+		strcpy(phrase, ",Y=?");
 	    else {
 		int used;
-		sprintf(reply + strlen(reply), ",Y=%d:", ud->satellites);
-		if (SEEN(ud->satellite_stamp))
+		sprintf(phrase, ",Y=%d:", ud->satellites);
+		if (SEEN(ud->satellite_stamp)) {
+		    int reported = 0;
 		    for (i = 0; i < ud->satellites; i++) {
 			used = 0;
 			for (j = 0; j < ud->satellites_used; j++)
@@ -284,20 +336,70 @@ static int handle_request(int fd, char *buf, int buflen)
 				used = 1;
 				break;
 			    }
-			if (ud->PRN[i])
-			    sprintf(reply + strlen(reply), "%d %d %d %d %d:", 
+			if (ud->PRN[i]) {
+			    sprintf(phrase+strlen(phrase), "%d %d %d %d %d:", 
 				    ud->PRN[i], 
 				    ud->elevation[i],ud->azimuth[i],
 				    ud->ss[i],
 				    used);
+			    reported++;
+			}
 		    }
+		    assert(reported == ud->satellites);
 		}
+	    }
 	    break;
+#ifdef PROFILING
+	case 'Z':
+	    if (*p == '=') ++p;
+	    if (*p == '1' || *p == '+') {
+		ud->profiling = 1;
+		gpsd_report(3, "%d turned on profiling mode\n", fd);
+		sprintf(phrase, ",Z=1");
+		p++;
+	    } else if (*p == '0' || *p == '-') {
+		ud->profiling = 0;
+		gpsd_report(3, "%d turned off profiling mode\n", fd);
+		sprintf(phrase, ",Z=0");
+		p++;
+	    } else if (FD_ISSET(fd, &nmea_fds)) {
+		ud->profiling = 0;
+		gpsd_report(3, "%d turned off profiling mode\n", fd);
+		sprintf(phrase, ",Z=0");
+	    } else {
+		ud->profiling = !ud->profiling;
+		gpsd_report(3, "%d toggled profiling mode\n", fd);
+		sprintf(phrase, ",Z=%d", ud->profiling);
+	    }
+	    break;
+#endif /* PROFILING */
+
 	case '\r': case '\n':
 	    goto breakout;
 	}
+	if (strlen(reply) + strlen(phrase) < sizeof(reply) - 1)
+	    strcat(reply, phrase);
+	else
+	    return -1;	/* Buffer would overflow.  Just return an error */
     }
  breakout:
+#ifdef PROFILING
+    if (ud->profiling && icd) {
+	struct timeval tv;
+	gettimeofday(&tv, NULL);
+	sprintf(phrase, ",$=%s %.4d %.4f %.4f %.4f %.4f %.4f %.4lf",
+		ud->tag,
+		ud->sentence_length,
+		ud->gps_time,
+		ud->d_xmit_time - ud->gps_time,
+		ud->d_recv_time - ud->gps_time,
+		ud->d_decode_time - ud->gps_time,
+		session->poll_times[fd] - ud->gps_time,
+		TIME2DOUBLE(tv) - ud->gps_time); 
+	if (strlen(reply) + strlen(phrase) < sizeof(reply) - 1)
+	    strcat(reply, phrase);
+    }
+#endif /* PROFILING */
     strcat(reply, "\r\n");
 
     return throttled_write(fd, reply, strlen(reply));
@@ -315,9 +417,22 @@ static void notify_watchers(char *sentence)
 
 static void raw_hook(char *sentence)
 /* hook to be executed on each incoming sentence */
+/* CAUTION: only one NMEA sentence per call */
 {
     int fd;
 
+#ifdef PROFILING
+    char *sp, *tp;
+    if (sentence[0] != '$')
+	session->gNMEAdata.tag[0] = '\0';
+    else {
+	for (tp = session->gNMEAdata.tag, sp = sentence+1; *sp && *sp != ','; sp++, tp++)
+	    *tp = *sp;
+	*tp = '\0';
+    }
+    session->gNMEAdata.sentence_length = strlen(sentence);
+#endif /* PROFILING */
+
     for (fd = 0; fd < nfds; fd++) {
 	/* copy raw NMEA sentences from GPS to clients in raw mode */
 	if (FD_ISSET(fd, &nmea_fds))
@@ -327,9 +442,9 @@ static void raw_hook(char *sentence)
 	if (FD_ISSET(fd, &watcher_fds)) {
 #define PUBLISH(fd, cmds)	handle_request(fd, cmds, sizeof(cmds)-1)
 	    if (PREFIX("$GPRMC", sentence)) {
-		PUBLISH(fd, "pdtvs");
+		PUBLISH(fd, "pdtvsm");
 	    } else if (PREFIX("$GPGGA", sentence)) {
-		PUBLISH(fd, "pdas");	
+		PUBLISH(fd, "pdasm");	
 	    } else if (PREFIX("$GPGLL", sentence)) {
 		PUBLISH(fd, "pd");
 	    } else if (PREFIX("$GPVTG", sentence)) {
@@ -399,8 +514,8 @@ int main(int argc, char *argv[])
     int option, msock, fd, need_gps; 
     extern char *optarg;
 
-    debuglevel = 1;
-    while ((option = getopt(argc, argv, "D:S:d:hnp:s:v"
+    debuglevel = 0;
+    while ((option = getopt(argc, argv, "D:S:d:hNnp:P:s:v"
 #if TRIPMATE_ENABLE || defined(ZODIAC_ENABLE)
 			    "i:"
 #endif /* TRIPMATE_ENABLE || defined(ZODIAC_ENABLE) */
@@ -416,6 +531,11 @@ int main(int argc, char *argv[])
 #endif /* NON_NMEA_ENABLE */
 	case 'D':
 	    debuglevel = (int) strtol(optarg, 0, 0);
+	    if (debuglevel >= 2)
+		go_background = 0;
+	    break;
+	case 'N':
+	    go_background = 0;
 	    break;
 	case 'S':
 	    service = optarg;
@@ -453,6 +573,9 @@ int main(int argc, char *argv[])
 	case 'p':
 	    device_name = optarg;
 	    break;
+	case 'P':
+	    pid_file = optarg;
+	    break;
 	case 's':
 	    gpsd_speed = atoi(optarg);
 	    break;
@@ -468,7 +591,7 @@ int main(int argc, char *argv[])
 
     if (!service)
 	service = getservbyname("gpsd", "tcp") ? "gpsd" : DEFAULT_GPSD_PORT;
-    if (debuglevel < 2)
+    if (go_background)
 	daemonize();
 
     /* Handle some signals */
@@ -494,11 +617,11 @@ int main(int argc, char *argv[])
 
     FD_ZERO(&all_fds); FD_ZERO(&nmea_fds); FD_ZERO(&watcher_fds);
     FD_SET(msock, &all_fds);
-    nfds = getdtablesize();
+    nfds = FD_SETSIZE;
 
     session = gpsd_init(gpstype, dgpsserver);
     if (gpsd_speed)
-	session->baudrate = gpsd_speed;
+	session->gNMEAdata.baudrate = gpsd_speed;
     session->gpsd_device = device_name;
     session->gNMEAdata.raw_hook = raw_hook;
     if (session->dsock >= 0)
@@ -569,7 +692,7 @@ int main(int argc, char *argv[])
 
 	/* accept and execute commands for all clients */
 	need_gps = 0;
-	for (fd = 0; fd < getdtablesize(); fd++) {
+	for (fd = 0; fd < nfds; fd++) {
 	    if (fd == msock || fd == session->gNMEAdata.gps_fd)
 		continue;
 	    /*
@@ -577,9 +700,6 @@ int main(int argc, char *argv[])
 	     * streaming (raw or watcher mode).
 	     */
 	    if (FD_ISSET(fd, &rfds) || FD_ISSET(fd, &nmea_fds) || FD_ISSET(fd, &watcher_fds)) {
-		char buf[BUFSIZE];
-		int buflen;
-
 		if (session->gNMEAdata.gps_fd == -1) {
 		    gpsd_deactivate(session);
 		    if (gpsd_activate(session) >= 0) {
@@ -588,16 +708,23 @@ int main(int argc, char *argv[])
 		}
 
 		if (FD_ISSET(fd, &rfds)) {
-		    buflen = read(fd, buf, sizeof(buf) - 1);
-		    if (buflen <= 0) {
-			(void) close(fd);
-			FD_CLR(fd, &all_fds);
-		    }
-		    buf[buflen] = '\0';
-		    gpsd_report(1, "<= client: %s", buf);
-		    if (handle_request(fd, buf, buflen) < 0) {
+		    char buf[BUFSIZ];
+		    int buflen;
+		    gpsd_report(3, "checking %d \n", fd);
+		    if ((buflen = read(fd, buf, sizeof(buf) - 1)) <= 0) {
 			(void) close(fd);
 			FD_CLR(fd, &all_fds);
+			FD_CLR(fd, &nmea_fds);
+			FD_CLR(fd, &watcher_fds);
+		    } else {
+		        buf[buflen] = '\0';
+			gpsd_report(1, "<= client: %s", buf);
+			if (handle_request(fd, buf, buflen) < 0) {
+			    (void) close(fd);
+			    FD_CLR(fd, &all_fds);
+			    FD_CLR(fd, &nmea_fds);
+			    FD_CLR(fd, &watcher_fds);
+			}
 		    }
 		}
 	    }
@@ -613,7 +740,5 @@ int main(int argc, char *argv[])
     }
 
     gpsd_wrap(session);
+    return 0;
 }
-
-
-
