@@ -1,9 +1,27 @@
+/* AIX requires this to be the first thing in the file.  */
+#ifndef __GNUC__
+# if HAVE_ALLOCA_H
+#  include <alloca.h>
+# else
+#  ifdef _AIX
+ #pragma alloca
+#  else
+#   ifndef alloca /* predefined by HP cc +Olibcalls */
+char *alloca ();
+#   endif
+#  endif
+# endif
+#endif
+
 #include <stdio.h>
 #include <unistd.h>
 #include <stdlib.h>
 #include <math.h>
 #include <string.h>
 #include <stdarg.h>
+#include <ctype.h>
+#define __USE_XOPEN
+#include <time.h>
 
 #include "gpsd.h"
 
@@ -13,51 +31,32 @@
  *
  **************************************************************************/
 
-static char *field(char *sentence, short n)
-/* return the nth comma-delimited field from the sentence */
-{
-    static char result[100];
-    char c, *p = sentence;
-    unsigned int i;
-
-    while (n-- > 0)
-        while ((c = *p++) != ',' && c != '\0')
-	    continue;
-    strncpy(result, p, sizeof(result)-1);
-    p = result;
-    i = 0;
-    while (*p && *p != ',' && *p != '*' && *p != '\r' && ++i < sizeof(result))
-	p++;
-    *p = '\0';
-    return result;
-}
-
-static void do_lat_lon(char *sentence, int begin, struct gps_data_t *out)
+static void do_lat_lon(char *field[], struct gps_data_t *out)
 /* process a pair of latitude/longitude fields starting at field index BEGIN */
 {
     double lat, lon, d, m;
     char str[20], *p;
     int updated = 0;
 
-    if (*(p = field(sentence, begin + 0)) != '\0') {
+    if (*(p = field[0]) != '\0') {
 	strncpy(str, p, 20);
 	sscanf(p, "%lf", &lat);
 	m = 100.0 * modf(lat / 100.0, &d);
 	lat = d + m / 60.0;
-	p = field(sentence, begin + 1);
+	p = field[1];
 	if (*p == 'S')
 	    lat = -lat;
 	if (out->latitude != lat)
 	    out->latitude = lat;
 	updated++;
     }
-    if (*(p = field(sentence, begin + 2)) != '\0') {
+    if (*(p = field[2]) != '\0') {
 	strncpy(str, p, 20);
 	sscanf(p, "%lf", &lon);
 	m = 100.0 * modf(lon / 100.0, &d);
 	lon = d + m / 60.0;
 
-	p = field(sentence, begin + 3);
+	p = field[3];
 	if (*p == 'W')
 	    lon = -lon;
 	if (out->longitude != lon)
@@ -69,24 +68,24 @@ static void do_lat_lon(char *sentence, int begin, struct gps_data_t *out)
     REFRESH(out->latlon_stamp);
 }
 
-static int update_field_i(char *sentence, int fld, int *dest)
+static int update_field_i(char *field, int *dest)
 /* update an integer-valued field */
 {
     int tmp, changed;
 
-    tmp = atoi(field(sentence, fld));
+    tmp = atoi(field);
     changed = (tmp != *dest);
     *dest = tmp;
     return changed;
 }
 
-static int update_field_f(char *sentence, int fld, double *dest)
+static int update_field_f(char *field, double *dest)
 /* update a float-valued field */
 {
     int changed;
     double tmp;
 
-    tmp = atof(field(sentence, fld));
+    tmp = atof(field);
     changed = (tmp != *dest);
     *dest = tmp;
     return changed;
@@ -148,13 +147,36 @@ static void merge_hhmmss(char *hhmmss, struct gps_data_t *out)
     strcat(out->utc, "Z");
 }
 
+#ifdef PROFILING
+static double iso8661_to_unix(char *isotime)
+{
+    char *dp = NULL;
+    struct tm tm;
+    double usec, res;
+    time_t now;
+
+    dp = strptime(isotime, "%Y-%m-%dT%H:%M:%S", &tm);
+    if (*dp == '.')
+	usec = strtod(dp, NULL);
+    else
+	usec = 0;
+    res = mktime(&tm) - timezone + usec;
+    now = time(NULL);
+    if (daylight && localtime(&now)->tm_isdst)
+	res -= 3600;
+    if (now != (int)res)
+	gpsd_report(4, "clock skew is %lf seconds\n", now-(int)res);
+    return res;
+}
+#endif /* PROFILING */
+
 /**************************************************************************
  *
  * NMEA sentence handling begins here
  *
  **************************************************************************/
 
-static void processGPRMC(char *sentence, struct gps_data_t *out)
+static void processGPRMC(int count, char *field[], struct gps_data_t *out)
 /* Recommend Minimum Specific GPS/TRANSIT Data */
 {
     /*
@@ -168,22 +190,53 @@ static void processGPRMC(char *sentence, struct gps_data_t *out)
            191194       Date of fix  19 November 1994
            020.3,E      Magnetic variation 20.3 deg East
 	   A            FAA mode indicator (NMEA 2.3 and later)
+                        A=autonomous, D=differential, E=Estimated,
+                        N=not valid, S=Simulator
            *68          mandatory nmea_checksum
 
      * SiRF chipsets don't return either Mode Indicator or magnetic variation.
      */
-    if (!strcmp(field(sentence, 2), "A")) {
-	merge_ddmmyy(field(sentence, 9), out);
-	merge_hhmmss(field(sentence, 1), out);
-	do_lat_lon(sentence, 3, out);
-	out->speed_stamp.changed = update_field_f(sentence, 7, &out->speed);
+    if (!strcmp(field[2], "A")) {
+	if (count > 9) {
+	    merge_ddmmyy(field[9], out);
+	    merge_hhmmss(field[1], out);
+    #ifdef PROFILING
+	    out->gps_time = iso8661_to_unix(out->utc);
+    #endif /* PROFILING */
+	}
+	do_lat_lon(&field[3], out);
+	out->speed_stamp.changed = update_field_f(field[7], &out->speed);
 	REFRESH(out->speed_stamp);
-	out->track_stamp.changed = update_field_f(sentence, 8, &out->track);
+	out->track_stamp.changed = update_field_f(field[8], &out->track);
 	REFRESH(out->track_stamp);
+	/*
+	 * This copes with GPSes like the Magellan EC-10X that *only* emit
+	 * GPRMC. In this case we set mode and status here so the client
+	 * code that relies on them won't mistakenly believe it has never
+	 * received a fix.
+	 */
+	if (!(out->seen_sentences & GPGGA) && out->status == STATUS_NO_FIX) {
+	    /* Upgrade to STATUS_FIX
+	     * Do not touch otherwise, may be STATUS_FIX or
+	     * STATUS_DGPS_FIX, cannot tell apart here
+	     */
+	    out->status_stamp.changed = 1;
+	    out->status = STATUS_FIX;
+	    REFRESH(out->status_stamp);
+	}
+	if (!(out->seen_sentences & GPGSA && out->mode <= MODE_NO_FIX)) {
+	    /* Upgrade to MODE_2D
+	     * Do not touch otherwise, may be MODE_3D or
+	     * MODE_3D, cannot tell apart here
+	     */
+	    out->mode_stamp.changed = 1;
+	    out->mode = MODE_2D;
+	    REFRESH(out->mode_stamp);
+	}
     }
 }
 
-static void processGPGLL(char *sentence, struct gps_data_t *out)
+static void processGPGLL(int count, char *field[], struct gps_data_t *out)
 /* Geographic position - Latitude, Longitude */
 {
     /* Introduced in NMEA 3.0.  Here are the fields:
@@ -205,15 +258,18 @@ static void processGPGLL(char *sentence, struct gps_data_t *out)
      * SiRF chipsets don't return the Mode Indicator.
      * This code copes gracefully with both quirks.
      */
-    char *status = field(sentence, 7);
+    char *status = field[7];
 
-    if (!strcmp(field(sentence, 6), "A") && status[0] != 'N') {
+    if (!strcmp(field[6], "A") && (count < 8 || *status != 'N')) {
 	int newstatus = out->status;
 
-	do_lat_lon(sentence, 1, out);
 	fake_mmddyyyy(out);
-	merge_hhmmss(field(sentence, 5), out);
-	if (status[0] == 'D')
+	merge_hhmmss(field[5], out);
+#ifdef PROFILING
+	out->gps_time = iso8661_to_unix(out->utc);
+#endif /* PROFILING */
+	do_lat_lon(&field[1], out);
+	if (count >= 8 && *status == 'D')
 	    newstatus = STATUS_DGPS_FIX;	/* differential */
 	else
 	    newstatus = STATUS_FIX;
@@ -224,7 +280,7 @@ static void processGPGLL(char *sentence, struct gps_data_t *out)
     }
 }
 
-static void processGPVTG(char *sentence, struct gps_data_t *out)
+static void processGPVTG(int c UNUSED, char *field[], struct gps_data_t *out)
 /* Track Made Good and Ground Speed */
 {
     /* There are two variants of GPVTG.  One looks like this:
@@ -250,16 +306,16 @@ static void processGPVTG(char *sentence, struct gps_data_t *out)
 
      * which means we want to extract field 5.  We cope with both.
      */
-    out->track_stamp.changed = update_field_f(sentence, 1, &out->track);;
+    out->track_stamp.changed = update_field_f(field[1], &out->track);;
     REFRESH(out->track_stamp);
-    if (field(sentence, 2)[0] == 'T')
-	out->speed_stamp.changed = update_field_f(sentence, 5, &out->speed);
+    if (field[2][0] == 'T')
+	out->speed_stamp.changed = update_field_f(field[5], &out->speed);
     else
-	out->speed_stamp.changed = update_field_f(sentence, 3, &out->speed);
+	out->speed_stamp.changed = update_field_f(field[3], &out->speed);
     REFRESH(out->speed_stamp);
 }
 
-static void processGPGGA(char *sentence, struct gps_data_t *out)
+static void processGPGGA(int c UNUSED, char *field[], struct gps_data_t *out)
 /* Global Positioning System Fix Data */
 {
     /*
@@ -276,17 +332,20 @@ static void processGPGGA(char *sentence, struct gps_data_t *out)
            (empty field) time in seconds since last DGPS update
            (empty field) DGPS station ID number (0000-1023)
     */
-    out->status_stamp.changed = update_field_i(sentence, 6, &out->status);
+    out->status_stamp.changed = update_field_i(field[6], &out->status);
     REFRESH(out->status_stamp);
     gpsd_report(3, "GPGGA sets status %d\n", out->status);
     if (out->status > STATUS_NO_FIX) {
 	char	*altitude;
 
 	fake_mmddyyyy(out);
-	merge_hhmmss(field(sentence, 1), out);
-	do_lat_lon(sentence, 2, out);
-        out->satellites_used = atoi(field(sentence, 7));
-	altitude = field(sentence, 9);
+	merge_hhmmss(field[1], out);
+#ifdef PROFILING
+	out->gps_time = iso8661_to_unix(out->utc);
+#endif /* PROFILING */
+	do_lat_lon(&field[2], out);
+        out->satellites_used = atoi(field[7]);
+	altitude = field[9];
 	/*
 	 * SiRF chipsets up to version 2.2 report a null altitude field.
 	 * See <http://www.sirf.com/Downloads/Technical/apnt0033.pdf>.
@@ -308,7 +367,7 @@ static void processGPGGA(char *sentence, struct gps_data_t *out)
     }
 }
 
-static void processGPGSA(char *sentence, struct gps_data_t *out)
+static void processGPGSA(int c UNUSED, char *field[], struct gps_data_t *out)
 /* GPS DOP and Active Satellites */
 {
     /*
@@ -325,19 +384,21 @@ static void processGPGSA(char *sentence, struct gps_data_t *out)
      */
     int i, changed = 0;
     
-    out->mode_stamp.changed = update_field_i(sentence, 2, &out->mode);
+    out->mode_stamp.changed = update_field_i(field[2], &out->mode);
     REFRESH(out->mode_stamp);
     gpsd_report(3, "GPGSA sets mode %d\n", out->mode);
-    changed |= update_field_f(sentence, 15, &out->pdop);
-    changed |= update_field_f(sentence, 16, &out->hdop);
-    changed |= update_field_f(sentence, 17, &out->vdop);
+    changed |= update_field_f(field[15], &out->pdop);
+    changed |= update_field_f(field[16], &out->hdop);
+    changed |= update_field_f(field[17], &out->vdop);
     for (i = 0; i < MAXCHANNELS; i++)
 	out->used[i] = 0;
     out->satellites_used = 0;
     for (i = 0; i < MAXCHANNELS; i++) {
-        out->used[i] = atoi(field(sentence, i+3));
-        if (out->used[i] > 0)
-                out->satellites_used++;
+        int prn = atoi(field[i+3]);
+        if (prn > 0) {
+           out->used[out->satellites_used] = prn;
+           out->satellites_used++;
+       }
     }
     out->fix_quality_stamp.changed = changed;
     REFRESH(out->fix_quality_stamp);
@@ -365,7 +426,7 @@ int nmea_sane_satellites(struct gps_data_t *out)
     return 0;
 }
 
-static void processGPGSV(char *sentence, struct gps_data_t *out)
+static void processGPGSV(int count, char *field[], struct gps_data_t *out)
 /* GPS Satellites in View */
 {
     /*
@@ -380,47 +441,42 @@ static void processGPGSV(char *sentence, struct gps_data_t *out)
            <repeat for up to 4 satellites per sentence>
                 There my be up to three GSV sentences in a data packet
      */
-    int changed, lower, upper, fldnum = 4;
-
-    out->await = atoi(field(sentence, 1));
-    if (sscanf(field(sentence, 2), "%d", &out->part) < 1)
+    int changed, fldnum;
+    if (count <= 3)
         return;
+    out->await = atoi(field[1]);
+    if (sscanf(field[2], "%d", &out->part) < 1)
+        return;
+    else if (out->part == 1)
+    {
+	memset(out->PRN,       '\0', sizeof(out->PRN));
+	memset(out->elevation, '\0', sizeof(out->elevation));
+	memset(out->azimuth,   '\0', sizeof(out->azimuth));
+	memset(out->ss,        '\0', sizeof(out->ss));
+	out->satellites = 0;
+    }
 
-    changed = update_field_i(sentence, 3, &out->satellites);
-    if (out->satellites > MAXCHANNELS) out->satellites = MAXCHANNELS;
-
-    lower = (out->part - 1) * 4;
-    upper = lower + 4;
-    while (lower < out->satellites && lower < upper) {
-	changed |= update_field_i(sentence, fldnum++, &out->PRN[lower]);
-	changed |= update_field_i(sentence, fldnum++, &out->elevation[lower]);
-	changed |= update_field_i(sentence, fldnum++, &out->azimuth[lower]);
-	if (*(field(sentence, fldnum)))
-	    changed |= update_field_i(sentence, fldnum, &out->ss[lower]);
-	fldnum++; lower++;
+    changed = 0;
+    for (fldnum = 4; fldnum < count; out->satellites++) {
+	changed |= update_field_i(field[fldnum++], &out->PRN[out->satellites]);
+	changed |= update_field_i(field[fldnum++], &out->elevation[out->satellites]);
+	changed |= update_field_i(field[fldnum++], &out->azimuth[out->satellites]);
+	changed |= update_field_i(field[fldnum++], &out->ss[out->satellites]);
     }
 
     /* not valid data until we've seen a complete set of parts */
     if (out->part < out->await)
 	gpsd_report(3, "Partial satellite data (%d of %d).\n", out->part, out->await);
+    else if (!nmea_sane_satellites(out))
+	gpsd_report(3, "Satellite data no good.\n");
     else {
-	/* trim off PRNs with spurious data attached */
-	while (out->satellites
-		    && !out->elevation[out->satellites-1]
-		    && !out->azimuth[out->satellites-1]
-		    && !out->ss[out->satellites-1])
-	    out->satellites--;
-	if (nmea_sane_satellites(out)) {
-	    gpsd_report(3, "Satellite data OK.\n");
-	    out->satellite_stamp.changed = changed;
-	    REFRESH(out->satellite_stamp);
-	}
-	else
-	    gpsd_report(3, "Satellite data no good.\n");
+	gpsd_report(3, "Satellite data OK.\n");
+	out->satellite_stamp.changed = changed;
+	REFRESH(out->satellite_stamp);
     }
 }
 
-static short nmea_checksum(char *sentence)
+static short nmea_checksum(char *sentence, unsigned char *correct_sum)
 /* is the checksum on the specified sentence good? */
 {
     unsigned char sum = '\0';
@@ -428,8 +484,10 @@ static short nmea_checksum(char *sentence)
 
     while ((c = *p++) != '*' && c != '\0')
 	sum ^= c;
+    if (correct_sum)
+        *correct_sum = sum;
     sprintf(csum, "%02X", sum);
-    return (strncmp(csum, p, 2) == 0);
+    return(toupper(csum[0])==toupper(p[0]))&&(toupper(csum[1])==toupper(p[1]));
 }
 
 /**************************************************************************
@@ -439,55 +497,149 @@ static short nmea_checksum(char *sentence)
  **************************************************************************/
 
 void nmea_add_checksum(char *sentence)
-/* add NMEA to a *-terminated sentence */
+/* add NMEA checksum to a *-terminated sentence */
 {
     unsigned char sum = '\0';
     char c, *p = sentence;
 
+    if (*p == '$')
+	p++;
     while ((c = *p++) != '*' && c != '\0')
 	sum ^= c;
+    if (c != '*')
+	*p++ = '*';
     sprintf(p, "%02X\r\n", sum);
 }
 
 int nmea_parse(char *sentence, struct gps_data_t *outdata)
 /* parse an NMEA sentence, unpack it into a session structure */
 {
-    if (nmea_checksum(sentence+1)) {
-	if (PREFIX("$GPRMC", sentence)) {
-	    processGPRMC(sentence, outdata);
-	} else if (PREFIX("$GPGGA", sentence)) {
-	    processGPGGA(sentence, outdata);
-	} else if (PREFIX("$GPGLL", sentence)) {
-	    processGPGLL(sentence, outdata);
-	} else if (PREFIX("$GPVTG", sentence)) {
-	    processGPVTG(sentence, outdata);
-	} else if (PREFIX("$GPGSA", sentence)) {
-	    processGPGSA(sentence, outdata);
-	} else if (PREFIX("$GPGSV", sentence)) {
-	    processGPGSV(sentence, outdata);
-	} else if (PREFIX("$PRWIZCH", sentence)) {
-	    /* do nothing */;
-	} else
-	    return -1;
+    typedef void (*nmea_decoder)(int count, char *f[], struct gps_data_t *out);
+    static struct {
+	char *name;
+	int mask;
+	nmea_decoder decoder;
+    } nmea_phrase[] = {
+	{"GPRMB", 0,    	NULL},
+	{"GPRMC", GPRMC,	processGPRMC},
+	{"GPGGA", GPGGA,	processGPGGA},
+	{"GPGLL", GPGLL,	processGPGLL},
+	{"GPVTG", GPVTG,	processGPVTG},
+	{"GPGSA", GPGSA,	processGPGSA},
+	{"GPGSV", GPGSV,	processGPGSV},
+	{"PRWIZCH", 0,  	NULL},
+    };
+
+    int retval = -1;
+    unsigned int i;
+    int count;
+    unsigned char sum;
+    char *p, *s;
+    char *field[80];
+
+    if ( ! nmea_checksum(sentence+1, &sum)) {
+        gpsd_report(1, "Bad NMEA checksum: '%s' should be %02X\n",
+                   sentence, sum);
+        return 0;
     }
-    return 0;
+
+    /* make an editable copy of the sentence */
+#ifdef AC_FUNC_ALLOCA
+    s = alloca(strlen(sentence)+1);
+    strcpy(s, sentence);
+#else
+    s = strdup(sentence);
+#endif
+    /* discard the checksum part */
+    for (p = s; (*p != '*') && (*p >= ' '); ) ++p;
+    *p = '\0';
+    /* split sentence copy on commas, filling the field array */
+    for (count = 0, p = s; p != NULL && *p != 0; ++count, p = strchr(p, ',')) {
+	*p = 0;
+	field[count] = ++p;
+    }
+    /* dispatch on field zero, the sentence tag */
+    for (i = 0; i < sizeof(nmea_phrase)/sizeof(nmea_phrase[0]); ++i) {
+        if (!strcmp(nmea_phrase[i].name, field[0])) {
+	    if (nmea_phrase[i].decoder)
+		(nmea_phrase[i].decoder)(count, field, outdata);
+	    if (nmea_phrase[i].mask)
+		outdata->seen_sentences |= nmea_phrase[i].mask;
+	    retval = 0;
+	    break;
+	}
+    }
+#ifndef AC_FUNC_ALLOCA
+    free(s);
+#endif
+    return retval;
 }
 
-void nmea_send(int fd, const char *fmt, ... )
+int nmea_send(int fd, const char *fmt, ... )
 /* ship a command to the GPS, adding * and correct checksum */
 {
+    unsigned int status;
     char buf[BUFSIZ];
     va_list ap;
 
     va_start(ap, fmt) ;
-#ifdef HAVE_VSNPRINTF
-    vsnprintf(buf + strlen(buf), sizeof(buf)-strlen(buf), fmt, ap);
-#else
-    vsprintf(buf + strlen(buf), fmt, ap);
-#endif
+    vsnprintf(buf, sizeof(buf)-5, fmt, ap);
     va_end(ap);
     strcat(buf, "*");
-    nmea_add_checksum(buf + 1);
-    write(fd, buf, strlen(buf));
+    nmea_add_checksum(buf);
+    status = write(fd, buf, strlen(buf));
+    if (status == strlen(buf)) {
+	gpsd_report(2, "=> GPS: %s\n", buf);
+	return status;
+    } else {
+	gpsd_report(2, "=> GPS: %s FAILED\n", buf);
+	return -1;
+    }
 }
 
+int nmea_validate_buffer(char *buf, size_t n)
+/* does this buffer look like it contains valid NMEA? */
+{
+    char	*sp, csum[3];
+    unsigned int sum;
+
+    /*
+     * It's OK to have leading garbage, but not trailing garbage.
+     * Leading garbage may just mean the port hasn't settled yet
+     * after a baud rate change; ignore it.
+     */
+    for (sp = buf; sp < buf + n && !isprint(*sp); sp++)
+	continue;
+
+    /* If no valid NMEA in the read buffer, crap out */
+    if (!(sp = strstr(sp, "$GP"))) {
+	gpsd_report(4, "no NMEA in the buffer\n");
+	return 0;
+    }
+
+    /*
+     * Check to see if we actually have a valid NMEA packet here.
+     * Trailing garbage means that the data accidentally looked
+     * like NMEA or that old data that really was NMEA happened to
+     * be sitting in the TTY buffer unread, but the new data we
+     * read is not sentences.  Second case shouldn't happen,
+     * because we flush the buffer after each speed change, but
+     * welcome to serial-programming hell.
+     */
+    sum = 0;
+    for (++sp; *sp != '*' && *sp != '\0'; sp++) {
+	if (!isascii(*sp)) {
+	    gpsd_report(4, "trailing garbage in the buffer\n");
+	    return 0;
+	}
+	sum ^= *sp;
+    }
+    sprintf(csum, "%02X", sum);
+    if (*sp == '\0' 
+	    || toupper(csum[0])!=toupper(sp[1])
+	    || toupper(csum[1])!=toupper(sp[2])) {
+	gpsd_report(4, "checksum incorrect\n");
+	return 0;
+    }
+    return 1;
+}
